local LIP = nil
local LibDeflate = nil
local libSotA_Commands = nil

local POIS_core = {
	version = "0.07",
	scriptName = "Point of Interest System (POIS) Core Module",
	userSettings = {},
	pointLibraries = {}
}

POIS_core.loadSettings = function()
	
	-- ensure the user setting file exists.  this will create it if it doesnt.
	local file = io.open(ShroudLuaPath .. '/pois/user.ini', "r")
	if file == nil then 
		file = io.open(ShroudLuaPath .. '/pois/user.ini', "w")
		if file then 
			file:write()
			file:close() 
		end
	else
		file:close()
	end
	
	POIS_core.userSettings = LIP().load(ShroudLuaPath .. '/pois/user.ini')

	if (POIS_core.userSettings["settings"] == nil) then
		POIS_core.userSettings["settings"] = {
			height_offset = 2.0,
			scale = 1.0,
			default_color = "white"
		}		
		
		LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
	end

	if (POIS_core.userSettings["colors"] == nil or POIS_core.userSettings["settings"]["default_color"] == "FFFFFF") then
		
		POIS_core.userSettings["settings"]["default_color"] = "white"
		
		POIS_core.userSettings["colors"] = {
			white = "FFFFFF",
			green = "00FF00",
			red = "FF0000",
			blue = "00A1C9",
			orange = "FFA500",
			purple = "FF0AFF",
			yellow = "FFF019"
		}
		
		LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
		
		-- update colors from values to names
		if (POIS_core.userSettings["scenes"] != nil) then
			for sceneKey, sceneRawName in pairs(POIS_core.userSettings["scenes"]) do
				if (POIS_core.userSettings[sceneKey] != nil) then
					for pointKey, pointDescription in pairs(POIS_core.userSettings[sceneKey]) do
						if (POIS_core.userSettings[sceneKey .. "_" .. pointKey]) then
							for colorName, colorValue in pairs(POIS_core.userSettings["colors"]) do
								if (POIS_core.userSettings[sceneKey .. "_" .. pointKey]["color"] == colorValue) then
									POIS_core.userSettings[sceneKey .. "_" .. pointKey]["color"] = colorName
									break
								end
							end
						end
					end
				end
			end
			
			LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
		end		
	end

	-- add in the view distance entry if it is missing
	if (POIS_core.userSettings["settings"]["view_distance"] == nil) then
		-- default is 0 (infinite)
		POIS_core.userSettings["settings"]["view_distance"] = 0
		LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
	end
	
	SES.invoke("POIS.onUserSettingsLoad", POIS_core.userSettings)
end

POIS_core.SetViewDistance = function(newDistance)

	if (tonumber(newDistance) < 0 or tonumber(newDistance) > 70) then 
		newDistance = 0
	end

	local oldDistance = POIS_core.userSettings["settings"]["view_distance"]
	POIS_core.userSettings["settings"]["view_distance"] = newDistance

	LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)

	SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, POIS_core.GetSceneKey(ShroudGetCurrentSceneNameRaw()))
	SES.invoke("POIS.onViewDistanceChange", oldDistance, newDistance)
end

POIS_core.SetScale = function(newScale)

	if (tonumber(newScale) < 0 or tonumber(newScale) > 3) then 
		newScale = 1
	end

	local oldScale = POIS_core.userSettings["settings"]["scale"]
	POIS_core.userSettings["settings"]["scale"] = newScale

	LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
	
	SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, POIS_core.GetSceneKey(ShroudGetCurrentSceneNameRaw()))
	SES.invoke("POIS.onScaleChange", oldScale, newScale)
end

POIS_core.SetHeightOffset = function(newOffset)

	if (tonumber(newOffset) < 0 or tonumber(newOffset) > 5) then 
		newOffset = 2
	end

	local oldOffset = POIS_core.userSettings["settings"]["height_offset"]
	POIS_core.userSettings["settings"]["height_offset"] = newOffset

	LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
	
	SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, POIS_core.GetSceneKey(ShroudGetCurrentSceneNameRaw()))
	SES.invoke("POIS.onHeightOffsetChange", oldOffset, newOffset)
end

POIS_core.ChangeColor = function(color, value)
	value = string.upper(value)
	value = string.gsub(value, "#", "")

	local oldColor = nil

	if (POIS_core.userSettings["colors"] != nil) then
		if (POIS_core.userSettings["colors"][color] != nil) then
			oldColor = POIS_core.userSettings["colors"][color]
			POIS_core.userSettings["colors"][color] = value
			LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
			SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, POIS_core.GetSceneKey(ShroudGetCurrentSceneNameRaw()))
			SES.invoke("POIS.onColorValueChange", color, oldColor, value)
		end
	end
end

POIS_core.GetSceneKey = function(rawSceneName) 
	rawSceneName = rawSceneName or ShroudGetCurrentSceneNameRaw()
	
	local targetKey = nil

	if (POIS_core.userSettings["scenes"] != nil) then 
		for sceneKey, sceneName in pairs(POIS_core.userSettings["scenes"]) do
			if (sceneName == rawSceneName) then 
				targetKey = sceneKey
				break
			end
		end
	end
	
	return targetKey
end

POIS_core.GetPointsByRawSceneName = function(rawSceneName)
	rawSceneName = rawSceneName or ShroudGetCurrentSceneNameRaw()

	local sceneKey = POIS_core.GetSceneKey(rawSceneName)

	return POIS_core.GetPoints(sceneKey)
end

POIS_core.SearchPoints = function(searchText, sceneKey, skipNotification)

	local foundPoints = {}

	if (sceneKey == nil) then
		for sceneKey, sceneRawName in pairs(POIS_core.userSettings["scenes"]) do
			local found = POIS_core.SearchPoints(searchText, sceneKey, true)			
			if (found[sceneKey] != nil) then
				foundPoints[sceneKey] = found[sceneKey]
			end
		end
		
		SES.invoke("POIS.onFoundPoints", foundPoints, searchText)
		
		return foundPoints
	else

		local points = POIS_core.GetPoints(sceneKey)

		for pointKey, pointData in pairs(points) do
			local searchFind = pointData["description"]:lower():find(searchText:lower())
			if (searchFind) then 
				if (foundPoints[sceneKey] == nil) then foundPoints[sceneKey] = {} end
				foundPoints[sceneKey][pointKey] = pointData["description"]
			end
		end		
		
		if (skipNotification == nil or skipNotification == false) then 
			SES.invoke("POIS.onFoundPoints", foundPoints, searchText)
		end
	end

	return foundPoints	
end

POIS_core.GetPoints = function(sceneKey)
	local points = {}

	if (sceneKey != nil) then 

		if POIS_core.userSettings[sceneKey] != nil then 
			for poiKey, poiDescription in pairs(POIS_core.userSettings[sceneKey]) do 		
				points[sceneKey .. "_" .. poiKey] = {
					description = poiDescription,
					sceneKey = sceneKey,
					poiKey = poiKey
				}
				for poiValueKey, poiValue in pairs(POIS_core.userSettings[sceneKey .. "_" .. poiKey]) do
					points[sceneKey .. "_" .. poiKey][poiValueKey] = poiValue
				end
			end
		end
	end
	
	return points
end

POIS_core.SetColor = function(pointKey, color)

	if (string.lower(pointKey) == "all") then
		local points = POIS_core.GetPointsByRawSceneName(ShroudGetCurrentSceneNameRaw())

		for poiKey, poiDescription in pairs(points) do
			POIS_core.SetColor(poiKey, color)
		end
	else
		local keys = POIS_core.utility.SplitPointKey(pointKey)
				
		if (keys.sceneKey != nil and keys.poiKey != nil) then
			if (keys.sceneKey != nil and POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey] != nil) then 

				POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey]["color"] = color;
				SES.invoke("POIS.onColorChange", keys.sceneKey .. "_" .. keys.poiKey, POIS_core.userSettings[keys.sceneKey][keys.poiKey], POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey])
				
				LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
				SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, keys.sceneKey)
			end	
		end
	end
end

POIS_core.SetVisibleByColor = function(color, visible)
	local points = POIS_core.GetPointsByRawSceneName(ShroudGetCurrentSceneNameRaw())

	for poiKey, poiDescription in pairs(points) do
		if (POIS_core.userSettings[poiKey]["color"] == color) then
			POIS_core.SetVisible(poiKey, visible)
		end
	end
end

POIS_core.SetVisible = function(pointKey, visible)

	pointKey = string.upper(pointKey)

	if (string.lower(pointKey) == "all") then
		local points = POIS_core.GetPointsByRawSceneName(ShroudGetCurrentSceneNameRaw())

		for poiKey, poiDescription in pairs(points) do
			POIS_core.SetVisible(poiKey, visible)
		end
	else
		local keys = POIS_core.utility.SplitPointKey(pointKey)
				
		if (keys.sceneKey != nil and keys.poiKey != nil) then
			if (keys.sceneKey != nil and POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey] != nil) then 

				if (visible == true) then 
					POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey]["show"] = "1";
					SES.invoke("POIS.onPointShow", keys.sceneKey .. "_" .. keys.poiKey, POIS_core.userSettings[keys.sceneKey][keys.poiKey], POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey])
				else
					POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey]["show"] = "0";
					SES.invoke("POIS.onPointHide", keys.sceneKey .. "_" .. keys.poiKey, POIS_core.userSettings[keys.sceneKey][keys.poiKey], POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey])
				end
				
				LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
				SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, keys.sceneKey)
			end	
		end
	end
end

POIS_core.AddScene = function(rawSceneName, sceneName)
	local sceneKey = POIS_core.GetSceneKey(rawSceneName)

	if (sceneKey == nil) then 
		sceneKey = POIS_core.utility.GetNextKey("S", POIS_core.userSettings["scenes"])
		if (POIS_core.userSettings["scenes"] == nil) then POIS_core.userSettings["scenes"] = {} end
		POIS_core.userSettings["scenes"][sceneKey] = rawSceneName
				
		if (POIS_core.userSettings["scene-names"] == nil) then POIS_core.userSettings["scene-names"] = {} end
		POIS_core.userSettings["scene-names"][sceneKey] = sceneName

		LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
		SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, sceneKey)
		SES.invoke("POIS.onSceneAdd", sceneKey, POIS_core.userSettings["scenes"][sceneKey])		
	end
		
	return sceneKey
end

POIS_core.AddPoint = function(pointName, rawSceneName, sceneName)

	local sceneKey = POIS_core.AddScene(rawSceneName, sceneName)
	local location = string.format("%.3f", ShroudPlayerX) .. "|" .. string.format("%.3f", ShroudPlayerY) .. "|" .. string.format("%.3f", ShroudPlayerZ)

	return POIS_core.AddPointData(sceneKey, pointName, location, "1", "1", POIS_core.userSettings["settings"]["default_color"])
end

POIS_core.AddPointData = function(sceneKey, pointName, location, show, visited, color)
	local poiKey = POIS_core.utility.GetNextKey("C", POIS_core.userSettings[sceneKey])

	if (POIS_core.userSettings[sceneKey] == nil) then POIS_core.userSettings[sceneKey] = {} end
	if (pointName == "") then pointName = "Point " .. sceneKey .. "_" .. poiKey end

	POIS_core.userSettings[sceneKey][poiKey] = pointName

	POIS_core.userSettings[sceneKey .. "_" .. poiKey] = {
		location = location,
		show = show,
		visited = visited,
		color = color
	}
	
	LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
	SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, sceneKey)
	SES.invoke("POIS.onPointAdd", sceneKey .. "_" .. poiKey, pointName, POIS_core.userSettings[sceneKey .. "_" .. poiKey])

	return sceneKey .. "_" .. poiKey
end


POIS_core.RemovePoint = function(pointKey)
	
	local keys = POIS_core.utility.SplitPointKey(pointKey)
	
	if (keys.sceneKey != nil and keys.poiKey != nil) then
		SES.invoke("POIS.onPointRemove", keys.sceneKey .. "_" .. keys.poiKey, POIS_core.userSettings[keys.sceneKey][keys.poiKey], POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey])
		
		POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey] = nil
		POIS_core.userSettings[keys.sceneKey][keys.poiKey] = nil
		
		LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
		SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, keys.sceneKey)
	end
end

POIS_core.RenamePoint = function(pointKey, newName)
	
	local keys = POIS_core.utility.SplitPointKey(pointKey)
	
	if (keys.sceneKey != nil and keys.poiKey != nil) then
				
		if (POIS_core.userSettings[keys.sceneKey][keys.poiKey] != nil) then
			local oldName = POIS_core.userSettings[keys.sceneKey][keys.poiKey]
			POIS_core.userSettings[keys.sceneKey][keys.poiKey] = newName

			SES.invoke("POIS.onPointRename", pointKey, oldName, newName)
			
			LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
			SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, keys.sceneKey)
		end
	end
end

POIS_core.ListPoints = function(rawSceneName)
	rawSceneName = rawSceneName or ShroudGetCurrentSceneNameRaw()
	
	local points = POIS_core.GetPointsByRawSceneName(rawSceneName)
	
	SES.invoke("POIS.onListPoints", rawSceneName, points)
	
	return points
end

POIS_core.ListScenes = function()

	local scenes = {}

	for sceneKey, sceneRawName in pairs (POIS_core.userSettings["scenes"]) do
		scenes[sceneKey] = {}
		scenes[sceneKey]["sceneRawName"] = sceneRawName
		scenes[sceneKey]["sceneName"] = POIS_core.userSettings["scene-names"][sceneKey]
	end

	SES.invoke("POIS.onListScenes", scenes)
	
	return scenes
end

POIS_core.ExportPoint = function(pointKey)
	local exportTable = {}
	local keys = POIS_core.utility.SplitPointKey(pointKey)
	
	if (keys.sceneKey != nil and keys.poiKey != nil) then
	
		-- create the scenes entry
		exportTable["scenes"] = {}
		exportTable["scenes"][keys.sceneKey] = POIS_core.userSettings["scenes"][keys.sceneKey]

		-- create the scene-names entry
		exportTable["scene-names"] = {}
		exportTable["scene-names"][keys.sceneKey] = POIS_core.userSettings["scene-names"][keys.sceneKey]

		-- create the point entry
		exportTable[keys.sceneKey] = {}
		exportTable[keys.sceneKey][keys.poiKey] = POIS_core.userSettings[keys.sceneKey][keys.poiKey]
		
		-- create the point entry
		exportTable[keys.sceneKey .. "_" .. keys.poiKey] = {}
		
		for pointDataKey, pointDataValue in pairs(POIS_core.userSettings[keys.sceneKey .. "_" .. keys.poiKey]) do
			exportTable[keys.sceneKey .. "_" .. keys.poiKey][pointDataKey] = pointDataValue
		end
				
		-- convert table to a string for compression
		local stringExportTable = POIS_core.utility.TableToString(exportTable)		
		local encodedExportTable = LibDeflate:EncodeForPrint(stringExportTable)

		SES.invoke("POIS.onPointExport", POIS_core.userSettings[keys.sceneKey][keys.poiKey], pointKey, encodedExportTable)
	end
end

POIS_core.ImportPoint = function(encodedExportTable)

	local stringExportTable = LibDeflate:DecodeForPrint(encodedExportTable)
	local exportTable = POIS_core.utility.StringToTable(stringExportTable)
	
	if (type(exportTable) == "table") then	
	
		-- check to see if the scene is already in the users settings.  if so, we need to use the current sceneKey.
		for exportSceneKey, exportSceneRawName in pairs(exportTable["scenes"]) do
			local importSceneKey = nil
			local sceneFound = false
		
			if (POIS_core.userSettings["scenes"] != nil) then 
				for sceneKey, sceneRawName in pairs(POIS_core.userSettings["scenes"]) do
					if (sceneRawName == exportSceneRawName) then
						importSceneKey = sceneKey
						sceneFound = true
						break
					end
				end
			end
			
			-- if we didnt find it, then add it
			if (sceneFound == false) then
				importSceneKey = POIS_core.AddScene(exportSceneRawName, exportTable["scene-names"][exportSceneKey])
				POIS_core.userSettings[importSceneKey] = {}
			end

			-- add the points from the scene
			for exportPointKey, exportPointName in pairs(exportTable[exportSceneKey]) do
			
				local currentPointFound = false
				
				-- see if a point already exists with that exact location
				if (POIS_core.userSettings[importSceneKey] != nil) then
					for pointKey, pointName in pairs(POIS_core.userSettings[importSceneKey]) do
						if (POIS_core.userSettings[importSceneKey .. "_" .. pointKey]["location"] == exportTable[exportSceneKey .. "_" .. exportPointKey]["location"]) then 
							currentPointFound = true
							break
						end
					end
				end
				
				if (currentPointFound == false) then				
					POIS_core.AddPointData(importSceneKey, exportPointName, exportTable[exportSceneKey .. "_" .. exportPointKey]["location"], exportTable[exportSceneKey .. "_" .. exportPointKey]["show"], exportTable[exportSceneKey .. "_" .. exportPointKey]["visited"], exportTable[exportSceneKey .. "_" .. exportPointKey]["color"])					
				end
			end
		end
		
		LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
		SES.invoke("POIS.onUserSettingsChange", POIS_core.userSettings, POIS_core.GetSceneKey(ShroudGetCurrentSceneNameRaw()))		
	end
end

-- optional discoverable distance parameter
POIS_core.ExportPoints = function(exportSceneKey)
	-- TODO: implement along with discoverable points
	--discoverableDistance = tonumber(discoverableDistance) or 0
	local discoverableDistance = 0
	
	local exportTable = {}
	
	for sceneKey, sceneRawName in pairs(POIS_core.userSettings["scenes"]) do
		if (exportSceneKey == nil or sceneKey == exportSceneKey) then
			-- create the scenes entry
			if exportTable["scenes"] == nil then exportTable["scenes"] = {} end
			exportTable["scenes"][sceneKey] = POIS_core.userSettings["scenes"][sceneKey]

			-- create the scene-names entry
			if exportTable["scene-names"] == nil then exportTable["scene-names"] = {} end
			exportTable["scene-names"][sceneKey] = POIS_core.userSettings["scene-names"][sceneKey]

			exportTable[sceneKey] = {}

			-- create the point entries
			for pointKey, pointName in pairs(POIS_core.userSettings[sceneKey]) do
						
				exportTable[sceneKey][pointKey] = POIS_core.userSettings[sceneKey][pointKey]
				
				-- create the point entry
				exportTable[sceneKey .. "_" .. pointKey] = {}
				
				exportTable[sceneKey .. "_" .. pointKey]["discoverable_distance"] = discoverableDistance
				
				for pointDataKey, pointDataValue in pairs(POIS_core.userSettings[sceneKey .. "_" .. pointKey]) do
				
					-- if this is discoverable, then initially hide
					if (discoverableDistance > 0 and pointDataKey == "show") then pointDataValue = "0" end
				
					exportTable[sceneKey .. "_" .. pointKey][pointDataKey] = pointDataValue
				end
			end
		end
	end
					
	-- convert table to a string for compression
	local stringExportTable = POIS_core.utility.TableToString(exportTable)		
	local encodedExportTable = LibDeflate:EncodeForPrint(stringExportTable)

	SES.invoke("POIS.onPointsExport", discoverableDistance, exportSceneKey, exportTable, encodedExportTable)
end

POIS_core.utility = {}

POIS_core.utility.SplitPointKey = function(pointKey)
	pointKey = string.upper(pointKey)

	local sceneKey = nil
	local poiKey = nil
	
	for key in string.gmatch(pointKey, "([^_]+)") do 
		if (sceneKey == nil) then
			sceneKey = key
		elseif (poiKey == nil) then
			poiKey = key
		end
	end
	
	return {sceneKey = sceneKey, poiKey = poiKey}
end

POIS_core.utility.GetNextKey = function(prefix, targetTable)
	local nextKey = nil
	local nextIndex = 0
	
	if (targetTable != nil) then 
		for key in pairs(targetTable) do
			nextIndex = key:gsub("[a-zA-z]", "")
		end
	end
	
	nextKey = prefix .. tostring(nextIndex + 1)
	
	return nextKey
end

POIS_core.utility.TableLength = function(targetTable)
	local length = 0
	
	for _ in pairs(targetTable) do
		length = length + 1
	end
	
	return length
end

POIS_core.utility.reconstruct = function(...)
	local paramValue = ""
	local previousValue = ""

	for i = select('#', ...), 1, -1 do
		local value = select(i, ...)
		
		if (value != nil) then 
			paramValue = value:gsub(previousValue, "") .. " " .. paramValue
			previousValue = value
		end
	end
	
	-- trim the return value
	return paramValue:gsub("^%s*(.-)%s*$", "%1")
end

-- Convert a lua table into a lua syntactically correct string and back again
-- https://gist.github.com/justnom/9816256
POIS_core.utility.TableToString = function(tbl)
    local result = "{"
    for k, v in pairs(tbl) do
        -- Check the key type (ignore any numerical keys - assume its an array)
        if type(k) == "string" then
            result = result.."[\""..k.."\"]".."="
        end

        -- Check the value type
        if type(v) == "table" then
            result = result .. POIS_core.utility.TableToString(v)
        elseif type(v) == "boolean" then
            result = result..tostring(v)
        else
            result = result.."\""..v.."\""
        end
        result = result..","
    end
    -- Remove leading commas from the result
    if result ~= "" then
        result = result:sub(1, result:len()-1)
    end
    return result.."}"
end

POIS_core.utility.StringToTable = function(stringtable)
	return loadsafe("return " .. stringtable)()
end


POIS_core.initializeCommands = function()
	libSotA_Commands.command.add("pois", function(source, action, ...)
			
		if (action == "list") then
			POIS_core.ListPoints()
		elseif (action == "list-scenes") then
			POIS_core.ListScenes()
		elseif (action == "show") then
			POIS_core.SetVisible(..., true)
		elseif (action == "hide") then
			POIS_core.SetVisible(..., false)
		elseif (action == "show-color") then
			POIS_core.SetVisibleByColor(..., true)
		elseif (action == "hide-color") then
			POIS_core.SetVisibleByColor(..., false)
		elseif (action == "add") then
			local pointName = POIS_core.utility.reconstruct(...)
			local sceneKey = POIS_core.AddPoint(pointName, ShroudGetCurrentSceneNameRaw(), ShroudGetCurrentSceneName())
		elseif (action == "remove") then
			POIS_core.RemovePoint(...)
		elseif (action == "set-view-distance") then
			POIS_core.SetViewDistance(...)
		elseif (action == "set-scale") then
			POIS_core.SetScale(...)
		elseif (action == "set-height-offset") then
			POIS_core.SetHeightOffset(...)
		elseif (string.find(action, "change%-color%-")) then
			local color = string.gsub(action, "change%-color%-", "")
			color = string.lower(color)
			POIS_core.ChangeColor(color, ...)
		elseif (string.find(action, "color-")) then
			local color = string.gsub(action, "color%-", "")
			color = string.lower(color)
			POIS_core.SetColor(..., color)
		elseif (action == "export-point") then
			POIS_core.ExportPoint(...)
		elseif (action == "import-point") then
			POIS_core.ImportPoint(...)
		elseif (action == "export-points") then
			POIS_core.ExportPoints(...)
		elseif (action == "rename-point") then
			local parameters = POIS_core.utility.reconstruct(...)
			
			local keyEnd = parameters:find(" ")
			
			local pointKey = parameters:sub(1, keyEnd - 1):upper()
			local newName = parameters:sub(keyEnd + 1)
			
			POIS_core.RenamePoint(pointKey, newName)
		elseif (action == "search-scene-points") then
			local searchText = POIS_core.utility.reconstruct(...)
			POIS_core.SearchPoints(searchText, POIS_core.GetSceneKey())
		elseif (action == "search-all-points") then
			local searchText = POIS_core.utility.reconstruct(...)
			POIS_core.SearchPoints(searchText)					
		elseif (action == "help") then
		
			local colorList = ""
			local colorCnt = 0
			
			for colorName, colorValue in pairs(POIS_core.userSettings["colors"]) do
				if (colorCnt > 0) then colorList = colorList .. ", " end
				colorList = colorList .. "[" .. colorValue .. "]" .. colorName .. "[-]"
				colorCnt = colorCnt + 1
			end
		
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("Commands are entered into local chat in the following format:")
			ConsoleLog("[FFFF00]!pois <command> <parameter>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("Available Commands")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To list points in the current scene:")
			ConsoleLog("[FFFF00]!pois list-points[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To list scenes with points:")
			ConsoleLog("[FFFF00]!pois list-scenes[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To add a point to the current scene:")
			ConsoleLog("[FFFF00]!pois add <point-name>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To rename a point in the current scene:")
			ConsoleLog("[FFFF00]!pois rename-point <point-key> <point-name>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To remove a point from the current scene:")
			ConsoleLog("[FFFF00]!pois remove <point-key>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To search for a point in the current scene:")
			ConsoleLog("[FFFF00]!pois search-scene-points <search-string>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To search for a point in all scenes:")
			ConsoleLog("[FFFF00]!pois search-all-points <search-string>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To show a point in the current scene:")
			ConsoleLog("[FFFF00]!pois show <point-key | all>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To hide a point in the current scene:")
			ConsoleLog("[FFFF00]!pois hide <point-key | all>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To show all points in the current scene of the same color:")
			ConsoleLog("[FFFF00]!pois show-color <color-name>[-]")
			ConsoleLog("Available colors: " .. colorList)
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To hide all points in the current scene of the same color:")
			ConsoleLog("[FFFF00]!pois hide-color <color-name>[-]")
			ConsoleLog("Available colors: " .. colorList)
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To change the color of a point in the current scene:")
			ConsoleLog("[FFFF00]!pois color-<color-name> <point-key>[-]")						
			ConsoleLog("Available colors: " .. colorList)
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To change a color:")
			ConsoleLog("[FFFF00]!pois change-color-<color-name> <color-value>[-]")
			ConsoleLog("Available colors: " .. colorList)
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To change distance points can be seen (0 for infinite):")
			ConsoleLog("[FFFF00]!pois set-view-distance <distance>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To change the scale of point labels (1 is default):")
			ConsoleLog("[FFFF00]!pois set-scale <scale>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To change the height offset of point labels (2 is default):")
			ConsoleLog("[FFFF00]!pois set-height-offset <offset>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To export a point:")
			ConsoleLog("[FFFF00]!pois export-point <point-key>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To import a point:")
			ConsoleLog("[FFFF00]!pois import-point <export-string>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To export a list of points:")
			ConsoleLog("[FFFF00]!pois export-points <scene-key>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
		end	
	end)
	
	libSotA_Commands.command.restrict("pois", "Local", ShroudGetPlayerName())
	
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Welcome to version " .. POIS_core.version .. " of the " .. POIS_core.scriptName .. " lua addon for Shroud of the Avatar")
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("This lua addon can be found on the SotA forum at [FFFF00]https://www.shroudoftheavatar.com/forum/index.php?threads/point-of-interest-system-addon-wip.172379/[-]")
	ConsoleLog("This lua addon utilizes the SotA Event System (SES) [FFFF00]https://www.shroudoftheavatar.com/forum/index.php?threads/sota-event-system-addon.172117/[-]")
	ConsoleLog("This lua addon utilizes a customized version of libsota.  Thank you to the original author CatweazleX.")
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Commands use the format [FFFF00]!pois <command> <parameter>[-] and are entered into local chat.")
	ConsoleLog("Use [FFFF00]!pois help[-] in local chat to see a list of available commands.")
	ConsoleLog("[808080]---------------------------------------------[-]")
	
end

POIS_core.onStart = function()
	LIP = SES.utility.loadmodule("pois/modules/LIP.lua")
	LibDeflate = SES.utility.loadmodule("pois/modules/LibDeflate")()
	libSotA_Commands = SES.utility.loadmodule("pois/modules/libsotamod/libsota-commands")()

	POIS_core.loadSettings()	
	POIS_core.initializeCommands()	
	
	-- kick off the current scene
	POIS_core.onSceneChange(ShroudGetCurrentSceneName(), ShroudGetCurrentSceneNameRaw())
end

POIS_core.onSceneChange = function(sceneName, sceneRawName)
	-- send out an event with sceneKey and points
	local sceneKey = POIS_core.GetSceneKey(sceneRawName)

	-- add the pretty scene name if it doesnt exist
	if (POIS_core.userSettings["scenes"] != nil and POIS_core.userSettings["scenes"][sceneKey] != nil) then
		if (POIS_core.userSettings["scene-names"] == nil) then POIS_core.userSettings["scene-names"] = {} end
		if (POIS_core.userSettings["scene-names"][sceneKey] == nil) then 
			POIS_core.userSettings["scene-names"][sceneKey] = sceneName
			LIP().save(ShroudLuaPath .. '/pois/user.ini', POIS_core.userSettings)
		end
	end
	
	SES.invoke("POIS.onSceneChange", sceneKey)
end

POIS_core.onListPoints = function(rawSceneName, points)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Points of Interest located in [FFFF00]" .. rawSceneName .. "[-]")
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("[808080]Point Key | Description | Visible[-]")
	ConsoleLog("[808080]---------------------------------------------[-]")

	for poiKey, poiData in pairs(points) do
		ConsoleLog("[FFFF00]" .. poiKey .. "[-] | [" .. POIS_core.userSettings["colors"][poiData["color"]] .. "]" .. poiData["description"] .. "[-] | " .. poiData["show"])
	end	
	
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onListScenes = function(scenes)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Scenes with Points of Interest")
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("[808080]Scene Key | Scene Name[-]")
	ConsoleLog("[808080]---------------------------------------------[-]")

	for sceneKey, sceneData in pairs(scenes) do
		ConsoleLog("[FFA500]" .. sceneKey .. "[-] | [FFFF00]" .. sceneData["sceneName"] .. "[-]")
	end	
	
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onSceneAdd = function(sceneKey, rawSceneName)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Scene [FFFF00]" .. rawSceneName .. "[-] ([FFA500]" .. sceneKey .. "[-]) has been [00FF00]added[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onPointAdd = function(pointKey, pointName, pointData)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Point of Interest [FFFF00]" .. pointName .. "[-] ([FFA500]" .. pointKey .. "[-]) has been [00FF00]added[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onPointRemove = function(pointKey, pointName, pointData)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Point of Interest [FFFF00]" .. pointName .. "[-] ([FFA500]" .. pointKey .. "[-]) has been [FF0000]removed[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onPointRename = function(pointKey, oldName, newName)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Point of Interest [FFFF00]" .. oldName .. "[-] ([FFA500]" .. pointKey .. "[-]) has been renamed to [FFFF00]" .. newName .. "[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onPointShow = function(pointKey, pointName, pointData)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Point of Interest [FFFF00]" .. pointName .. "[-] ([FFA500]" .. pointKey .. "[-]) is [00FF00]showing[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onPointHide = function(pointKey, pointName, pointData)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Point of Interest [FFFF00]" .. pointName .. "[-] ([FFA500]" .. pointKey .. "[-]) is [FF0000]hidden[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onColorChange = function(pointKey, pointName, pointData)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Point of Interest [FFFF00]" .. pointName .. "[-] ([FFA500]" .. pointKey .. "[-]) has been changed to [" .. POIS_core.userSettings["colors"][pointData["color"]] .. "]" .. pointData["color"] .. "[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onColorValueChange = function(color, oldValue, newValue)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Color [FFFF00]" .. color .. "[-] has been changed from [" .. oldValue .. "]" .. oldValue .. "[-] to [" .. newValue .. "]" .. newValue .. "[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onViewDistanceChange = function(oldValue, newValue)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("View distance has been changed from " .. oldValue .. " to " .. newValue .. ".")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onScaleChange = function(oldValue, newValue)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Scale has been changed from " .. oldValue .. " to " .. newValue .. ".")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onHeightOffsetChange = function(oldOffset, newOffset)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Height offset has been changed from " .. oldOffset .. " to " .. newOffset .. ".")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

POIS_core.onPointExport = function(pointName, pointKey, encodedExportTable)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Point of Interest [FFFF00]" .. pointName .. "[-] ([FFA500]" .. pointKey .. "[-]) has been exported to your ChatLog file.  Open your ChatLog and copy the text between the --- to use for import.")
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog(encodedExportTable)		
	ConsoleLog("[808080]---------------------------------------------[-]")		
end

POIS_core.onPointsExport = function(discoverableDistance, exportSceneKey, exportTable, encodedExportTable)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Points of Interest have been exported to your ChatLog file.  Open your ChatLog and copy the text between the --- to use for import or point library.")
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog(encodedExportTable)		
	ConsoleLog("[808080]---------------------------------------------[-]")			
end

POIS_core.onFoundPoints = function(foundPoints, searchText)

	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Point search results for [FFFF00]" .. searchText .. "[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")

	for sceneKey, pointList in pairs(foundPoints) do
		ConsoleLog("[808080]---------------------------------------------[-]")
		ConsoleLog("Points located in [FFFF00]" .. POIS_core.userSettings["scene-names"][sceneKey] .. "[-].")
		ConsoleLog("[808080]---------------------------------------------[-]")
		ConsoleLog("[808080]Point Key | Description[-]")
		ConsoleLog("[808080]---------------------------------------------[-]")
		for pointKey, pointDescription in pairs(pointList) do
			ConsoleLog("[FFFF00]" .. pointKey .. "[-] | [" .. POIS_core.userSettings["colors"][POIS_core.userSettings[pointKey]["color"]] .. "]" .. pointDescription .. "[-]")
		end
	end
end

local loadLibraryQueue = {}
local librariesToLoad = false

POIS_core.onLoadLibrary = function(libraryKey, scriptName, version, creatorName, description, encodedExportTable)
	LibDeflate = SES.utility.loadmodule("pois/modules/LibDeflate")()

	local decodedExportTable = LibDeflate:DecodeForPrint(encodedExportTable)
	local libraryPointData = POIS_core.utility.StringToTable(decodedExportTable)		

	loadLibraryQueue[libraryKey] = {}
	loadLibraryQueue[libraryKey]["data"] = libraryPointData
	loadLibraryQueue[libraryKey]["scriptName"] = scriptName
	loadLibraryQueue[libraryKey]["version"] = version
	loadLibraryQueue[libraryKey]["creatorName"] = creatorName
	loadLibraryQueue[libraryKey]["description"] = description	

	librariesToLoad = true
		
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Point library " .. scriptName .. " v" .. version .. " created by " .. creatorName .. " has been loaded.")
	ConsoleLog("[808080]---------------------------------------------[-]")	
end

POIS_core.onUnloadLibrary = function(libraryKey)
	
	if (POIS_core.pointLibraries[libraryKey] != nil) then
		local scriptName = POIS_core.pointLibraries[libraryKey]["scriptName"]
		
		POIS_core.pointLibraries[libraryKey] = nil
			
		ConsoleLog("[808080]---------------------------------------------[-]")
		ConsoleLog("Point library " .. scriptName .. " has been removed.")
		ConsoleLog("[808080]---------------------------------------------[-]")
		
		SES.invoke("POIS.onLibraryChange", POIS_core.pointLibraries)
	end
end

POIS_core.onUpdate = function()

	if (librariesToLoad == true) then 
		for libraryKey, libraryData in pairs(loadLibraryQueue) do
			POIS_core.pointLibraries[libraryKey] = libraryData		
		end

		loadLibraryQueue = {}
		librariesToLoad = false

		SES.invoke("POIS.onLibraryChange", POIS_core.pointLibraries)
	end
end

function POIS_core:initialize()

	SES.on("ShroudOnStart", POIS_core.onStart)
	SES.on("ShroudOnUpdate", POIS_core.onUpdate)
	SES.on("ShroudOnSceneChange", POIS_core.onSceneChange)
	SES.on("POIS.onListPoints", POIS_core.onListPoints)
	SES.on("POIS.onListScenes", POIS_core.onListScenes)	
	SES.on("POIS.onSceneAdd", POIS_core.onSceneAdd)
	SES.on("POIS.onPointAdd", POIS_core.onPointAdd)
	SES.on("POIS.onPointRemove", POIS_core.onPointRemove)
	SES.on("POIS.onPointRename", POIS_core.onPointRename)
	SES.on("POIS.onPointShow", POIS_core.onPointShow)
	SES.on("POIS.onPointHide", POIS_core.onPointHide)
	SES.on("POIS.onColorChange", POIS_core.onColorChange)
	SES.on("POIS.onColorValueChange", POIS_core.onColorValueChange)
	SES.on("POIS.onViewDistanceChange", POIS_core.onViewDistanceChange)
	SES.on("POIS.onScaleChange", POIS_core.onScaleChange)
	SES.on("POIS.onHeightOffsetChange", POIS_core.onHeightOffsetChange)	
	SES.on("POIS.onPointExport", POIS_core.onPointExport)
	SES.on("POIS.onPointsExport", POIS_core.onPointsExport)
	SES.on("POIS.onFoundPoints", POIS_core.onFoundPoints)
	SES.on("POIS.onLoadLibrary", POIS_core.onLoadLibrary)
	SES.on("POIS.onUnloadLibrary", POIS_core.onUnloadLibrary)

	return self
end

return POIS_core:initialize()