
local POIS_ui = {
	version = "0.07",
	scriptName = "Point of Interest System (POIS) Display Module",
	refreshInterval = 0.01,
	userSettings = nil,
	pointLibraries = nil,
	currentSceneKey = nil,
	currentScenePoints = nil
}

POIS_ui.utility = {}

POIS_ui.GetLibrarySceneKey = function(libraryKey, rawSceneName) 
	rawSceneName = rawSceneName or ShroudGetCurrentSceneNameRaw()
	
	local targetKey = nil

	if (POIS_ui.pointLibraries != nil and POIS_ui.pointLibraries[libraryKey] != nil) then
		for sceneKey, sceneRawName in pairs (POIS_ui.pointLibraries[libraryKey]["data"]["scenes"]) do
			if (sceneRawName == rawSceneName) then 
				targetKey = sceneKey
				break
			end
		end
	end
	return targetKey
end

POIS_ui.GetVisiblePoints = function(sceneKey)
	local points = {}

	if (sceneKey != nil) then
		if POIS_ui.userSettings[sceneKey] != nil then 
			for poiKey, poiDescription in pairs(POIS_ui.userSettings[sceneKey]) do 
				if (tostring(POIS_ui.userSettings[sceneKey .. "_" .. poiKey]["show"]) == "1") then 				
					points[sceneKey .. "_" .. poiKey] = {
						description = poiDescription,
						sceneKey = sceneKey,
						poiKey = poiKey,
						displayID = -1
					}
					for poiValueKey, poiValue in pairs(POIS_ui.userSettings[sceneKey .. "_" .. poiKey]) do
						points[sceneKey .. "_" .. poiKey][poiValueKey] = poiValue
					end
				end
			end
		end
	end
	
	if (POIS_ui.pointLibraries != nil) then
		for pointLibraryKey, pointLibraryData in pairs(POIS_ui.pointLibraries) do
			local librarySceneKey = POIS_ui.GetLibrarySceneKey(pointLibraryKey)
			if pointLibraryData["data"][librarySceneKey] != nil then
				for poiKey, poiDescription in pairs(pointLibraryData["data"][librarySceneKey]) do 	
					if (tostring(pointLibraryData["data"][librarySceneKey .. "_" .. poiKey]["show"]) == "1") then 				
						points[librarySceneKey .. "_PL_" .. poiKey] = {
							description = poiDescription,
							libraryKey = pointLibraryKey,
							sceneKey = librarySceneKey,
							poiKey = poiKey,
							displayID = -1
						}
						for poiValueKey, poiValue in pairs(pointLibraryData["data"][librarySceneKey .. "_" .. poiKey]) do
							points[librarySceneKey .. "_PL_" .. poiKey][poiValueKey] = poiValue
						end
					end
				end
			end
		end
	end		
	
	return points
end

POIS_ui.UpdateCurrentPoints = function()

	if (POIS_ui.currentScenePoints != nil) then
		-- if we have points, destroy the display objects	
		for pointKey, pointData in pairs(POIS_ui.currentScenePoints) do
			if (pointData["displayID"] >= 0) then
				ShroudDestroyObject(pointData["displayID"], UI.Text)
				POIS_ui.currentScenePoints[pointKey]["displayID"] = -1
			end
		end
	end
	
	POIS_ui.currentScenePoints = POIS_ui.GetVisiblePoints(POIS_ui.currentSceneKey)
	
	-- create new display objects
	for pointKey, pointData in pairs(POIS_ui.currentScenePoints) do
		if (pointData["displayID"] == -1) then
			local displayID = ShroudUIText(pointData["description"], 20, -100, 0, 200, 50)
			ShroudSetScale(displayID, UI.Text, tonumber(POIS_ui.userSettings["settings"]["scale"]))
			ShroudSetColor(displayID, UI.Text, "#" .. POIS_ui.userSettings["colors"][pointData["color"]])
			
			POIS_ui.currentScenePoints[pointKey]["displayID"] = displayID
		end
	end
end

local uiIsActive = true
POIS_ui.onUIActiveChange = function(isActive)
	uiIsActive = isActive
end

POIS_ui.onSceneChange = function(sceneKey)
	POIS_ui.currentSceneKey = sceneKey
	POIS_ui.UpdateCurrentPoints()
end

POIS_ui.onSettingsChange = function(newUserSettings, currentSceneKey)
	POIS_ui.currentSceneKey = currentSceneKey
	POIS_ui.userSettings = newUserSettings
	POIS_ui.UpdateCurrentPoints()
end

POIS_ui.onLibraryChange = function(pointLibraries)
	POIS_ui.pointLibraries = pointLibraries
	POIS_ui.UpdateCurrentPoints()
end

POIS_ui.refreshDisplay = function()
	if (POIS_ui.currentScenePoints == nil) then return end
	
	for pointKey, pointData in pairs(POIS_ui.currentScenePoints) do
		if (pointData != nil) then
			local location = POIS_ui.utility.SplitLocation(pointData["location"])
									
			local screenPoint = nil
			
			if (type(ShroudWorldToScreenPoint) == "nil") then
				screenPoint = {x = 200, y = 500, z = 10}
			else
				screenPoint = ShroudWorldToScreenPoint(location.x, location.y + POIS_ui.userSettings["settings"]["height_offset"], location.z)
			end

			if (screenPoint.z < 0 or uiIsActive == false) then 
				ShroudHideObject(pointData["displayID"], UI.Text)
			else
				local distance = screenPoint.z
				
				if (tonumber(POIS_ui.userSettings["settings"]["view_distance"]) > 0 and tonumber(POIS_ui.userSettings["settings"]["view_distance"]) < distance) then
					ShroudHideObject(pointData["displayID"], UI.Text)
				else
					
					if distance > 70 then distance = 70 end
					local distanceScaling = (1 - distance / 100)
				
					ShroudSetScale(pointData["displayID"], UI.Text, POIS_ui.userSettings["settings"]["scale"] * distanceScaling)
					ShroudShowObject(pointData["displayID"], UI.Text)
					ShroudSetPosition(pointData["displayID"], UI.Text, screenPoint.x, ShroudGetScreenY() - screenPoint.y)
				end
			end
		end
	end
end

POIS_ui.utility.SplitLocation = function(location)
	local locationX = nil
	local locationY = nil
	local locationZ = nil
	
	for value in string.gmatch(location, "([^|]+)") do 
		if (value != nil) then 
			if (locationX == nil) then
				locationX = tonumber(value)
			elseif (locationY == nil) then
				locationY = tonumber(value)
			elseif (locationZ == nil) then
				locationZ = tonumber(value)
			end
		end
	end
	
	return {x = locationX, y = locationY, z = locationZ}
end

local lastPoll = nil

POIS_ui.onUpdate = function()
	if not ShroudServerTime then return end -- shroud Api not ready, yet
	
	local ts = os.time()
	if (lastPoll != nil and (ts - lastPoll) < POIS_ui.refreshInterval) then return end

	POIS_ui.refreshDisplay()

	lastPoll = ts		
end

function POIS_ui:initialize()
	SES.on("ShroudOnUIActiveChange", POIS_ui.onUIActiveChange)
	SES.on("ShroudOnUpdate", POIS_ui.onUpdate)
	SES.on("POIS.onSceneChange", POIS_ui.onSceneChange)	
	SES.on("POIS.onUserSettingsLoad", POIS_ui.onSettingsChange)
	SES.on("POIS.onUserSettingsChange", POIS_ui.onSettingsChange)
	SES.on("POIS.onLibraryChange", POIS_ui.onLibraryChange)
	
	return self
end

return POIS_ui:initialize()