--[[
 * libsota-timers.lua
 * Copyright (C) 2022 Garritt Grandberg (aka Tirrag/peace-n-love)

 This file is a port of the timers functionality from libsota put into a module format.

   libsota is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   libsota is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.";
]]


-- TODO: validate for SES

local libsota_timers = {
}

libsota_timers.timer = {
	ui_timer_user_poll_interval = 0.120,
	list = {},
	add = function(timeout, once, callback, ...)
		local index = #libsota_timers.timer.list + 1
		local interval = nil

		if not once then
			interval = timeout
		end

		libsota_timers.timer.list[index] = {
			_index = index,
			time = os.time() + timeout,
			interval = interval,
			enabled = true,
			callback = callback,
			userdata = table.pack(...),
		}
		return index
	end,
	get = function(index)
		return libsota_timers.timer.list[index]
	end,
	remove = function(index)
		libsota_timers.timer.list[index] = nil
	end,
	enabled = function(index, enabled)
		if enabled ~= nil then
			libsota_timers.timer.list[index].enabled = enabled
		end
		return libsota_timers.timer.list[index].enabled
	end,
	pause = function(index)
		libsota_timers.timer.list[index].enabled = false
	end,
	resume = function(index)
		libsota_timers.timer.list[index].enabled = true
	end,
	toggle = function(index)
		libsota_timers.timer.list[index].enabled = not libsota_timers.timer.list[index].enabled
	end,
	setTimeout = function(timeout, callback, ...) return libsota_timers.timer.add(timeout, true, callback, ...) end,
	setInterval = function(interval, callback, ...) return libsota_timers.timer.add(interval, false, callback, ...) end,
	onStart = function() 
		-- TODO: add ConsoleLog info?
	end,
	ui_timer_user_last_ts = nil,
	onUpdate = function() 
		local ts = os.time()
				
		-- see if we need to run yet
		if (libsota_timers.timer.ui_timer_user_last_ts != nil and (ts - libsota_timers.timer.ui_timer_user_last_ts) < libsota_timers.timer.ui_timer_user_poll_interval) then return end

		for _,t in next, libsota_timers.timer.list do
			if t.enabled and ts >= t.time then
				if t.interval then
					t.time = ts + t.interval
				else
					libsota_timers.timer.list[t._index] = nil
				end
				if t.callback(unpack(t.userdata)) then libsota_timers.timer.list[t._index] = nil end
			end
		end
		
		libsota_timers.timer.ui_timer_user_last_ts = ts
	end,
}

function libsota_timers.timer:initialize()
	SES.on("ShroudOnStart", libsota_timers.timer.onStart)
	SES.on("ShroudOnUpdate", libsota_timers.timer.onUpdate)
	
	return self
end

return libsota_timers.timer:initialize()